/*
This file is part of the Juju GUI, which lets users view and manage Juju
environments within a graphical interface (https://launchpad.net/juju-gui).
Copyright (C) 2015 Canonical Ltd.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License version 3, as published by
the Free Software Foundation.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
General Public License for more details.

You should have received a copy of the GNU Affero General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

'use strict';

YUI.add('entity-content', function () {

  juju.components.EntityContent = React.createClass({
    /* Define and validate the properites available on this component. */
    propTypes: {
      apiUrl: React.PropTypes.string.isRequired,
      changeState: React.PropTypes.func.isRequired,
      entityModel: React.PropTypes.object.isRequired,
      getFile: React.PropTypes.func.isRequired,
      hasPlans: React.PropTypes.bool.isRequired,
      plans: React.PropTypes.array,
      pluralize: React.PropTypes.func.isRequired,
      renderMarkdown: React.PropTypes.func.isRequired
    },

    /**
      Generate the list of configuration options for a charm.
       @method _generateCharmConfig
      @param {Object} entityModel The entity model.
      @return {Object} The options markup.
    */
    _generateCharmConfig: function (entityModel) {
      var options = entityModel.get('options');
      if (!options) {
        return;
      }
      var optionsList = [];
      Object.keys(options).forEach(function (name) {
        var option = options[name];
        option.name = name;
        optionsList.push(React.createElement(juju.components.EntityContentConfigOption, {
          key: name,
          option: option }));
      }, this);
      return React.createElement(
        'dl',
        null,
        optionsList
      );
    },

    /**
      Generate the list of configuration options for a bundle.
       @method _generateBundleConfig
      @param {Object} entityModel The entity model.
      @return {Object} The options markup.
    */
    _generateBundleConfig: function (entityModel) {
      var services = entityModel.get('services');
      if (!services) {
        return;
      }
      // Generate the options for each service in this bundle.
      var servicesList = Object.keys(services).map(service => {
        var options = services[service].options || {};
        // Generate the list of options for this service.
        var optionsList = Object.keys(options).map((name, i) => {
          return React.createElement(
            'div',
            { className: 'entity-content__config-option',
              key: name + i },
            React.createElement(
              'dt',
              { className: 'entity-content__config-name' },
              name
            ),
            React.createElement(
              'dd',
              { className: 'entity-content__config-description' },
              React.createElement(
                'p',
                null,
                options[name]
              )
            )
          );
        });
        var classes = {
          'entity-content__bundle-config': true
        };
        if (optionsList.length === 0) {
          optionsList.push(React.createElement(
            'div',
            { key: 'none' },
            'No config options for this service.'
          ));
        }
        return React.createElement(
          juju.components.ExpandingRow,
          {
            classes: classes,
            key: service },
          React.createElement(
            'div',
            { className: 'entity-content__bundle-config-title' },
            service,
            React.createElement(
              'div',
              { className: 'entity-content__bundle-config-chevron' },
              React.createElement(
                'div',
                { className: 'entity-content__bundle-config-expand' },
                React.createElement(juju.components.SvgIcon, {
                  name: 'chevron_down_16',
                  size: '16' })
              ),
              React.createElement(
                'div',
                { className: 'entity-content__bundle-config-contract' },
                React.createElement(juju.components.SvgIcon, {
                  name: 'chevron_up_16',
                  size: '16' })
              )
            )
          ),
          React.createElement(
            'dl',
            { className: 'entity-content__bundle-config-options' },
            optionsList
          )
        );
      });
      return React.createElement(
        'ul',
        null,
        servicesList
      );
    },

    /**
      Generate the list of configuration options.
       @method _generateOptionsList
      @param {Object} entityModel The entity model.
      @return {Object} The options markup.
    */
    _generateOptionsList: function (entityModel) {
      var optionsList;
      if (entityModel.get('entityType') === 'charm') {
        optionsList = this._generateCharmConfig(entityModel);
      } else {
        optionsList = this._generateBundleConfig(entityModel);
      }
      if (optionsList) {
        return React.createElement(
          'div',
          { id: 'configuration',
            className: 'row row--grey entity-content__configuration' },
          React.createElement(
            'div',
            { className: 'inner-wrapper' },
            React.createElement(
              'div',
              { className: 'twelve-col' },
              React.createElement(
                'h2',
                { className: 'entity-content__header' },
                'Configuration'
              ),
              optionsList
            )
          )
        );
      }
    },

    /**
      Generates an HTML list from the supplied array.
       @method _generateList
      @param {Array} list The list of objects to markup.
      @param {Function} handler The click handler for each item.
      @return {Array} The list markup.
    */
    _generateList: function (list, handler) {
      return list.map(function (item, i) {
        return React.createElement(
          'li',
          { key: item + i },
          React.createElement(
            'a',
            { 'data-id': item, onClick: handler },
            item
          )
        );
      });
    },

    /**
      Generate the list of Tags if available.
       @method _generateTags
      @return {Array} The tags markup.
    */
    _generateTags: function () {
      // Have to convert {0: 'database'} to ['database'].
      var tags = [],
          entityTags = this.props.entityModel.get('tags'),
          index;
      if (!entityTags) {
        return;
      }
      for (index in entityTags) {
        tags.push(entityTags[index]);
      }
      return React.createElement(
        'div',
        { className: 'four-col entity-content__metadata' },
        React.createElement(
          'h4',
          null,
          'Tags'
        ),
        React.createElement(
          'ul',
          null,
          this._generateList(tags, this._handleTagClick)
        )
      );
    },

    /**
      Handle clicks on tags.
       @method _handleTagClick
      @param {Object} e The event.
    */
    _handleTagClick: function (e) {
      e.stopPropagation();
      this.props.changeState({
        sectionC: {
          component: 'charmbrowser',
          metadata: {
            activeComponent: 'search-results',
            search: null,
            tags: e.target.getAttribute('data-id')
          }
        }
      });
    },

    /**
      Generate the description if it is a charm.
       @method _generateDescription
      @param {Object} entityModel The entity model.
      @return {Object} The description markup.
    */
    _generateDescription: function (entityModel) {
      if (entityModel.get('entityType') === 'charm') {
        var bugLink = 'https://bugs.launchpad.net/charms/+source/' + `${ entityModel.get('name') }`;
        var submitLink = 'https://bugs.launchpad.net/charms/+source/' + `${ entityModel.get('name') }/+filebug`;
        var contributeLink = 'https://code.launchpad.net/~charmers/charms/' + `${ entityModel.get('series') }/${ entityModel.get('name') }/trunk`;
        return React.createElement(
          'div',
          { className: 'row row--grey entity-content__description' },
          React.createElement(
            'div',
            { className: 'inner-wrapper' },
            React.createElement(
              'div',
              { className: 'twelve-col' },
              React.createElement(
                'p',
                { className: 'intro' },
                entityModel.get('description')
              )
            ),
            this._generateTags(),
            React.createElement(
              'div',
              { className: 'four-col entity-content__metadata last-col' },
              React.createElement(
                'h4',
                null,
                'More information'
              ),
              React.createElement(
                'ul',
                null,
                React.createElement(
                  'li',
                  null,
                  React.createElement(
                    'a',
                    { href: bugLink, target: '_blank' },
                    'Bugs'
                  )
                ),
                React.createElement(
                  'li',
                  null,
                  React.createElement(
                    'a',
                    { href: submitLink, target: '_blank' },
                    'Submit a bug'
                  )
                ),
                React.createElement(
                  'li',
                  null,
                  React.createElement(
                    'a',
                    { href: contributeLink, target: '_blank' },
                    'Contribute'
                  )
                )
              )
            )
          )
        );
      }
    },

    /**
      Display the resources if this is a charm.
       @method _generateResources
    */
    _generateResources: function () {
      var entityModel = this.props.entityModel;
      if (entityModel.get('entityType') === 'charm') {
        return React.createElement(juju.components.EntityResources, {
          pluralize: this.props.pluralize,
          resources: entityModel.get('resources') });
      }
    },

    /**
      We only show the relations when it's a charm, but not a bundle.
       @method _showEntityRelations
    */
    _showEntityRelations: function () {
      var entityModel = this.props.entityModel;
      if (entityModel.get('entityType') === 'charm') {
        // Need to flatten out the relations to determine if we have any.
        var relations = entityModel.get('relations');
        // Normal behavior when there are no relations is to provide an empty
        // object. That said, in the interest of defensively protecting against
        // unexpected data, make sure these variables are at least empty
        // objects.
        var requires = relations.requires || {};
        var provides = relations.provides || {};
        var relationsList = Object.keys(requires).concat(Object.keys(provides));
        if (relationsList.length > 0) {
          return React.createElement(juju.components.EntityContentRelations, {
            changeState: this.props.changeState,
            relations: relations });
        }
      }
    },

    /**
      Generate the actions links.
       @method _generateActions
    */
    _generateActions: function () {
      var entityModel = this.props.entityModel;
      if (entityModel.get('entityType') !== 'bundle') {
        return;
      }
      var contributeLink = 'https://code.launchpad.net/~charmers/charms/' + `bundles/${ entityModel.get('name') }/bundle`;
      return React.createElement(
        'div',
        { className: 'section' },
        React.createElement(
          'h3',
          { className: 'section__title' },
          'Actions'
        ),
        React.createElement(
          'a',
          { href: contributeLink,
            target: '_blank' },
          'Contribute'
        )
      );
    },

    /**
      Transform and generate the price list from a provided string of prices.
       @method _generatePriceList
      @param {String} prices The string of prices in the format
        'price/quantity;price/quantity'.
      @returns {Object} The price list JSX components.
    */
    _generatePriceList: function (prices) {
      var prices = prices.split(';');
      var priceList = [];
      prices.forEach((price, i) => {
        if (price === '') {
          return;
        }
        var [amount, quantity] = price.split('/');
        var quantityItem;
        if (quantity) {
          quantityItem = React.createElement(
            'span',
            { className: 'entity-content__plan-price-quantity' },
            '/ ',
            quantity
          );
        }
        priceList.push(React.createElement(
          'li',
          { className: 'entity-content__plan-price-item',
            key: amount + (quantity || '') + i },
          React.createElement(
            'span',
            { className: 'entity-content__plan-price-amount' },
            amount
          ),
          quantityItem
        ));
      });
      return priceList;
    },

    /**
      Generate the list of plans.
       @method _generatePlans
    */
    _generatePlans: function () {
      var props = this.props;
      if (props.entityModel.get('entityType') !== 'charm' || !this.props.hasPlans) {
        return;
      }
      var plans = props.plans;
      // Return a spinner if null (we don't have a response yet) or nothing if
      // plans are a 0-length array (no plans found, likely due to an error).
      if (!plans) {
        return React.createElement(juju.components.Spinner, null);
      }
      if (!plans.length) {
        return;
      }
      var plansList = [];
      plans.forEach((plan, i) => {
        var classes = classNames('entity-content__plan', 'four-col', { 'last-col': (i + 1) % 3 === 0 });

        plansList.push(React.createElement(
          'div',
          { className: classes,
            key: plan.url + i },
          React.createElement(
            'div',
            { className: 'entity-content__plan-content' },
            React.createElement(
              'h3',
              { className: 'entity-content__plan-title' },
              plan.url
            ),
            React.createElement(
              'ul',
              { className: 'entity-content__plan-price' },
              this._generatePriceList(plan.price)
            ),
            React.createElement(
              'p',
              { className: 'entity-content__plan-description' },
              plan.description
            )
          )
        ));
      });
      return React.createElement(
        'div',
        { id: 'plans',
          className: 'row entity-content__plans' },
        React.createElement(
          'div',
          { className: 'inner-wrapper' },
          React.createElement(
            'div',
            { className: 'twelve-col' },
            React.createElement(
              'h2',
              { className: 'entity-content__header' },
              'Plans'
            ),
            React.createElement(
              'div',
              { className: 'equal-height' },
              plansList
            )
          )
        )
      );
    },

    render: function () {
      var entityModel = this.props.entityModel;
      return React.createElement(
        'div',
        { className: 'entity-content' },
        this._generateDescription(entityModel),
        this._generatePlans(),
        React.createElement(
          'div',
          { className: 'row' },
          React.createElement(
            'div',
            { className: 'inner-wrapper' },
            React.createElement(
              'div',
              { className: 'seven-col append-one' },
              React.createElement(juju.components.EntityContentReadme, {
                entityModel: entityModel,
                renderMarkdown: this.props.renderMarkdown,
                getFile: this.props.getFile })
            ),
            React.createElement(
              'div',
              { className: 'four-col' },
              this._generateResources(),
              this._showEntityRelations(),
              React.createElement(juju.components.EntityFiles, {
                apiUrl: this.props.apiUrl,
                entityModel: entityModel,
                pluralize: this.props.pluralize }),
              React.createElement(juju.components.EntityContentRevisions, {
                revisions: entityModel.get('revisions') }),
              this._generateActions()
            )
          )
        ),
        this._generateOptionsList(entityModel)
      );
    }
  });
}, '0.1.0', {
  requires: ['entity-content-config-option', 'entity-content-readme', 'entity-content-relations', 'entity-content-revisions', 'entity-files', 'entity-resources', 'expanding-row', 'loading-spinner', 'svg-icon']
});