/*
This file is part of the Juju GUI, which lets users view and manage Juju
environments within a graphical interface (https://launchpad.net/juju-gui).
Copyright (C) 2015 Canonical Ltd.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License version 3, as published by
the Free Software Foundation.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
General Public License for more details.

You should have received a copy of the GNU Affero General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

'use strict';

var juju = { components: {} }; // eslint-disable-line no-unused-vars

chai.config.includeStack = true;
chai.config.truncateThreshold = 0;

function _generateTagItem(tag, fn) {
  return [React.createElement(
    'li',
    { key: tag + 0 },
    React.createElement(
      'a',
      { 'data-id': tag, onClick: fn },
      tag
    )
  )];
}

describe('EntityContent', function () {
  var mockEntity;

  beforeAll(function (done) {
    // By loading these files it makes their classes available in the tests.
    YUI().use('entity-content', function () {
      done();
    });
  });

  beforeEach(function () {
    mockEntity = jsTestUtils.makeEntity();
  });

  afterEach(function () {
    mockEntity = undefined;
  });

  it('can display a charm', function () {
    var apiUrl = 'http://example.com';
    var renderMarkdown = sinon.spy();
    var getFile = sinon.spy();
    var changeState = sinon.spy();
    var pluralize = sinon.spy();
    mockEntity.set('resources', { resource: 'one' });
    var renderer = jsTestUtils.shallowRender(React.createElement(juju.components.EntityContent, {
      apiUrl: apiUrl,
      changeState: changeState,
      entityModel: mockEntity,
      getFile: getFile,
      hasPlans: false,
      pluralize: pluralize,
      renderMarkdown: renderMarkdown }), true);
    var option1 = {
      description: 'Your username',
      type: 'string',
      default: 'spinach',
      name: 'username'
    };
    var option2 = {
      description: 'Your password',
      type: 'string',
      default: 'abc123',
      name: 'password'
    };
    var instance = renderer.getMountedInstance();
    var output = renderer.getRenderOutput();
    var expected = React.createElement(
      'div',
      { className: 'entity-content' },
      React.createElement(
        'div',
        { className: 'row row--grey entity-content__description' },
        React.createElement(
          'div',
          { className: 'inner-wrapper' },
          React.createElement(
            'div',
            { className: 'twelve-col' },
            React.createElement(
              'p',
              { className: 'intro' },
              'Django framework.'
            )
          ),
          React.createElement(
            'div',
            { className: 'four-col entity-content__metadata' },
            React.createElement(
              'h4',
              null,
              'Tags'
            ),
            React.createElement(
              'ul',
              null,
              _generateTagItem('database', instance._handleTagClick)
            )
          ),
          React.createElement(
            'div',
            { className: 'four-col entity-content__metadata last-col' },
            React.createElement(
              'h4',
              null,
              'More information'
            ),
            React.createElement(
              'ul',
              null,
              React.createElement(
                'li',
                null,
                React.createElement(
                  'a',
                  {
                    href: 'https://bugs.launchpad.net/charms/+source/django',
                    target: '_blank' },
                  'Bugs'
                )
              ),
              React.createElement(
                'li',
                null,
                React.createElement(
                  'a',
                  { href: 'https://bugs.launchpad.net/charms/+source/' + 'django/+filebug', target: '_blank' },
                  'Submit a bug'
                )
              ),
              React.createElement(
                'li',
                null,
                React.createElement(
                  'a',
                  { href: 'https://code.launchpad.net/~charmers/charms/' + 'trusty/django/trunk', target: '_blank' },
                  'Contribute'
                )
              )
            )
          )
        )
      ),
      undefined,
      React.createElement(
        'div',
        { className: 'row' },
        React.createElement(
          'div',
          { className: 'inner-wrapper' },
          React.createElement(
            'div',
            { className: 'seven-col append-one' },
            React.createElement(juju.components.EntityContentReadme, {
              entityModel: mockEntity,
              renderMarkdown: renderMarkdown,
              getFile: getFile })
          ),
          React.createElement(
            'div',
            { className: 'four-col' },
            React.createElement(juju.components.EntityResources, {
              pluralize: pluralize,
              resources: { resource: 'one' } }),
            React.createElement(juju.components.EntityContentRelations, {
              changeState: changeState,
              relations: mockEntity.get('relations') }),
            React.createElement(juju.components.EntityFiles, {
              apiUrl: apiUrl,
              entityModel: mockEntity,
              pluralize: pluralize }),
            React.createElement(juju.components.EntityContentRevisions, {
              revisions: mockEntity.get('revisions') }),
            undefined
          )
        )
      ),
      React.createElement(
        'div',
        { id: 'configuration',
          className: 'row row--grey entity-content__configuration' },
        React.createElement(
          'div',
          { className: 'inner-wrapper' },
          React.createElement(
            'div',
            { className: 'twelve-col' },
            React.createElement(
              'h2',
              { className: 'entity-content__header' },
              'Configuration'
            ),
            React.createElement(
              'dl',
              null,
              React.createElement(juju.components.EntityContentConfigOption, {
                key: option1.name,
                option: option1 }),
              React.createElement(juju.components.EntityContentConfigOption, {
                key: option2.name,
                option: option2 })
            )
          )
        )
      )
    );
    assert.deepEqual(output, expected);
  });

  it('can display a charm with no options', function () {
    mockEntity.set('options', null);
    var apiUrl = 'http://example.com';
    var renderMarkdown = sinon.spy();
    var getFile = sinon.spy();
    var pluralize = sinon.spy();
    var changeState = sinon.spy();
    var renderer = jsTestUtils.shallowRender(React.createElement(juju.components.EntityContent, {
      apiUrl: apiUrl,
      changeState: changeState,
      entityModel: mockEntity,
      getFile: getFile,
      hasPlans: false,
      pluralize: pluralize,
      renderMarkdown: renderMarkdown }), true);
    var instance = renderer.getMountedInstance();
    var output = renderer.getRenderOutput();
    var expected = React.createElement(
      'div',
      { className: 'entity-content' },
      React.createElement(
        'div',
        { className: 'row row--grey entity-content__description' },
        React.createElement(
          'div',
          { className: 'inner-wrapper' },
          React.createElement(
            'div',
            { className: 'twelve-col' },
            React.createElement(
              'p',
              { className: 'intro' },
              'Django framework.'
            )
          ),
          React.createElement(
            'div',
            { className: 'four-col entity-content__metadata' },
            React.createElement(
              'h4',
              null,
              'Tags'
            ),
            React.createElement(
              'ul',
              null,
              _generateTagItem('database', instance._handleTagClick)
            )
          ),
          React.createElement(
            'div',
            { className: 'four-col entity-content__metadata last-col' },
            React.createElement(
              'h4',
              null,
              'More information'
            ),
            React.createElement(
              'ul',
              null,
              React.createElement(
                'li',
                null,
                React.createElement(
                  'a',
                  {
                    href: 'https://bugs.launchpad.net/charms/+source/django',
                    target: '_blank' },
                  'Bugs'
                )
              ),
              React.createElement(
                'li',
                null,
                React.createElement(
                  'a',
                  { href: 'https://bugs.launchpad.net/charms/+source/' + 'django/+filebug', target: '_blank' },
                  'Submit a bug'
                )
              ),
              React.createElement(
                'li',
                null,
                React.createElement(
                  'a',
                  { href: 'https://code.launchpad.net/~charmers/charms/' + 'trusty/django/trunk', target: '_blank' },
                  'Contribute'
                )
              )
            )
          )
        )
      ),
      undefined,
      React.createElement(
        'div',
        { className: 'row' },
        React.createElement(
          'div',
          { className: 'inner-wrapper' },
          React.createElement(
            'div',
            { className: 'seven-col append-one' },
            React.createElement(juju.components.EntityContentReadme, {
              entityModel: mockEntity,
              renderMarkdown: renderMarkdown,
              getFile: getFile })
          ),
          React.createElement(
            'div',
            { className: 'four-col' },
            React.createElement(juju.components.EntityResources, {
              pluralize: pluralize,
              resources: undefined }),
            React.createElement(juju.components.EntityContentRelations, {
              changeState: changeState,
              relations: mockEntity.get('relations') }),
            React.createElement(juju.components.EntityFiles, {
              apiUrl: apiUrl,
              entityModel: mockEntity,
              pluralize: pluralize }),
            React.createElement(juju.components.EntityContentRevisions, {
              revisions: mockEntity.get('revisions') }),
            undefined
          )
        )
      ),
      undefined
    );
    assert.deepEqual(output, expected);
  });

  it('can display a bundle', function () {
    var apiUrl = 'http://example.com';
    var renderMarkdown = sinon.spy();
    var getFile = sinon.spy();
    var changeState = sinon.spy();
    var pluralize = sinon.spy();
    var mockEntity = jsTestUtils.makeEntity(true);
    var output = jsTestUtils.shallowRender(React.createElement(juju.components.EntityContent, {
      apiUrl: apiUrl,
      changeState: changeState,
      entityModel: mockEntity,
      getFile: getFile,
      hasPlans: false,
      pluralize: pluralize,
      renderMarkdown: renderMarkdown }));
    var expected = React.createElement(
      'div',
      { className: 'entity-content' },
      undefined,
      undefined,
      React.createElement(
        'div',
        { className: 'row' },
        React.createElement(
          'div',
          { className: 'inner-wrapper' },
          React.createElement(
            'div',
            { className: 'seven-col append-one' },
            React.createElement(juju.components.EntityContentReadme, {
              entityModel: mockEntity,
              renderMarkdown: renderMarkdown,
              getFile: getFile })
          ),
          React.createElement(
            'div',
            { className: 'four-col' },
            undefined,
            undefined,
            React.createElement(juju.components.EntityFiles, {
              apiUrl: apiUrl,
              entityModel: mockEntity,
              pluralize: pluralize }),
            React.createElement(juju.components.EntityContentRevisions, {
              revisions: mockEntity.get('revisions') }),
            React.createElement(
              'div',
              { className: 'section' },
              React.createElement(
                'h3',
                { className: 'section__title' },
                'Actions'
              ),
              React.createElement(
                'a',
                { href: 'https://code.launchpad.net/~charmers/charms/' + 'bundles/django-cluster/bundle',
                  target: '_blank' },
                'Contribute'
              )
            )
          )
        )
      ),
      React.createElement(
        'div',
        { id: 'configuration',
          className: 'row row--grey entity-content__configuration' },
        React.createElement(
          'div',
          { className: 'inner-wrapper' },
          React.createElement(
            'div',
            { className: 'twelve-col' },
            React.createElement(
              'h2',
              { className: 'entity-content__header' },
              'Configuration'
            ),
            React.createElement(
              'ul',
              null,
              React.createElement(
                juju.components.ExpandingRow,
                {
                  classes: {
                    'entity-content__bundle-config': true
                  },
                  key: 'gunicorn' },
                React.createElement(
                  'div',
                  { className: 'entity-content__bundle-config-title' },
                  'gunicorn',
                  React.createElement(
                    'div',
                    { className: 'entity-content__bundle-config-chevron' },
                    React.createElement(
                      'div',
                      { className: 'entity-content__bundle-config-expand' },
                      React.createElement(juju.components.SvgIcon, {
                        name: 'chevron_down_16',
                        size: '16' })
                    ),
                    React.createElement(
                      'div',
                      { className: 'entity-content__bundle-config-contract' },
                      React.createElement(juju.components.SvgIcon, {
                        name: 'chevron_up_16',
                        size: '16' })
                    )
                  )
                ),
                React.createElement(
                  'dl',
                  { className: 'entity-content__bundle-config-options' },
                  [React.createElement(
                    'div',
                    { className: 'entity-content__config-option',
                      key: 'name0' },
                    React.createElement(
                      'dt',
                      { className: 'entity-content__config-name' },
                      'name'
                    ),
                    React.createElement(
                      'dd',
                      { className: 'entity-content__config-description' },
                      React.createElement(
                        'p',
                        null,
                        'title'
                      )
                    )
                  ), React.createElement(
                    'div',
                    { className: 'entity-content__config-option',
                      key: 'active1' },
                    React.createElement(
                      'dt',
                      { className: 'entity-content__config-name' },
                      'active'
                    ),
                    React.createElement(
                      'dd',
                      { className: 'entity-content__config-description' },
                      React.createElement(
                        'p',
                        null,
                        true
                      )
                    )
                  )]
                )
              ),
              React.createElement(
                juju.components.ExpandingRow,
                {
                  classes: {
                    'entity-content__bundle-config': true
                  },
                  key: 'django' },
                React.createElement(
                  'div',
                  { className: 'entity-content__bundle-config-title' },
                  'django',
                  React.createElement(
                    'div',
                    { className: 'entity-content__bundle-config-chevron' },
                    React.createElement(
                      'div',
                      { className: 'entity-content__bundle-config-expand' },
                      React.createElement(juju.components.SvgIcon, {
                        name: 'chevron_down_16',
                        size: '16' })
                    ),
                    React.createElement(
                      'div',
                      { className: 'entity-content__bundle-config-contract' },
                      React.createElement(juju.components.SvgIcon, {
                        name: 'chevron_up_16',
                        size: '16' })
                    )
                  )
                ),
                React.createElement(
                  'dl',
                  { className: 'entity-content__bundle-config-options' },
                  [React.createElement(
                    'div',
                    { key: 'none' },
                    'No config options for this service.'
                  )]
                )
              )
            )
          )
        )
      )
    );
    assert.deepEqual(output, expected);
  });

  it('doesn\'t show relations when they don\'t exist', function () {
    var apiUrl = 'http://example.com';
    var renderMarkdown = sinon.spy();
    var getFile = sinon.spy();
    var changeState = sinon.spy();
    var pluralize = sinon.spy();
    mockEntity.set('relations', { requires: {}, provides: {} });
    var renderer = jsTestUtils.shallowRender(React.createElement(juju.components.EntityContent, {
      apiUrl: apiUrl,
      changeState: changeState,
      entityModel: mockEntity,
      getFile: getFile,
      hasPlans: false,
      pluralize: pluralize,
      renderMarkdown: renderMarkdown }), true);
    var output = renderer.getRenderOutput();
    var parent = output.props.children[2].props.children.props.children[1];
    var relationsComponent = parent.props.children[1];
    assert.equal(relationsComponent, undefined);
  });

  it('can display plans', function () {
    var plans = [{
      url: 'plan1',
      price: 'test/price1',
      description: 'description1'
    }, {
      url: 'plan2',
      price: 'price2;',
      description: 'description2'
    }, {
      url: 'plan3',
      price: 'test/price3;price3b',
      description: 'description3'
    }];
    mockEntity.set('options', null);
    var apiUrl = 'http://example.com';
    var renderMarkdown = sinon.spy();
    var getFile = sinon.spy();
    var pluralize = sinon.spy();
    var changeState = sinon.spy();
    var renderer = jsTestUtils.shallowRender(React.createElement(juju.components.EntityContent, {
      apiUrl: apiUrl,
      changeState: changeState,
      entityModel: mockEntity,
      getFile: getFile,
      hasPlans: true,
      plans: plans,
      pluralize: pluralize,
      renderMarkdown: renderMarkdown }), true);
    var output = renderer.getRenderOutput();
    var expected = React.createElement(
      'div',
      { id: 'plans',
        className: 'row entity-content__plans' },
      React.createElement(
        'div',
        { className: 'inner-wrapper' },
        React.createElement(
          'div',
          { className: 'twelve-col' },
          React.createElement(
            'h2',
            { className: 'entity-content__header' },
            'Plans'
          ),
          React.createElement(
            'div',
            { className: 'equal-height' },
            [React.createElement(
              'div',
              { className: 'entity-content__plan four-col',
                key: 'plan10' },
              React.createElement(
                'div',
                { className: 'entity-content__plan-content' },
                React.createElement(
                  'h3',
                  { className: 'entity-content__plan-title' },
                  'plan1'
                ),
                React.createElement(
                  'ul',
                  { className: 'entity-content__plan-price' },
                  [React.createElement(
                    'li',
                    { className: 'entity-content__plan-price-item',
                      key: 'testprice10' },
                    React.createElement(
                      'span',
                      { className: 'entity-content__plan-price-amount' },
                      'test'
                    ),
                    React.createElement(
                      'span',
                      { className: 'entity-content__plan-price-quantity' },
                      '/ ',
                      'price1'
                    )
                  )]
                ),
                React.createElement(
                  'p',
                  { className: 'entity-content__plan-description' },
                  'description1'
                )
              )
            ), React.createElement(
              'div',
              { className: 'entity-content__plan four-col',
                key: 'plan21' },
              React.createElement(
                'div',
                { className: 'entity-content__plan-content' },
                React.createElement(
                  'h3',
                  { className: 'entity-content__plan-title' },
                  'plan2'
                ),
                React.createElement(
                  'ul',
                  { className: 'entity-content__plan-price' },
                  [React.createElement(
                    'li',
                    { className: 'entity-content__plan-price-item',
                      key: 'price20' },
                    React.createElement(
                      'span',
                      { className: 'entity-content__plan-price-amount' },
                      'price2'
                    ),
                    undefined
                  )]
                ),
                React.createElement(
                  'p',
                  { className: 'entity-content__plan-description' },
                  'description2'
                )
              )
            ), React.createElement(
              'div',
              { className: 'entity-content__plan four-col last-col',
                key: 'plan32' },
              React.createElement(
                'div',
                { className: 'entity-content__plan-content' },
                React.createElement(
                  'h3',
                  { className: 'entity-content__plan-title' },
                  'plan3'
                ),
                React.createElement(
                  'ul',
                  { className: 'entity-content__plan-price' },
                  React.createElement(
                    'li',
                    { className: 'entity-content__plan-price-item',
                      key: 'testprice30' },
                    React.createElement(
                      'span',
                      { className: 'entity-content__plan-price-amount' },
                      'test'
                    ),
                    React.createElement(
                      'span',
                      { className: 'entity-content__plan-price-quantity' },
                      '/ ',
                      'price3'
                    )
                  ),
                  React.createElement(
                    'li',
                    { className: 'entity-content__plan-price-item',
                      key: 'price3b1' },
                    React.createElement(
                      'span',
                      { className: 'entity-content__plan-price-amount' },
                      'price3b'
                    ),
                    undefined
                  )
                ),
                React.createElement(
                  'p',
                  { className: 'entity-content__plan-description' },
                  'description3'
                )
              )
            )]
          )
        )
      )
    );
    assert.deepEqual(output.props.children[1], expected);
  });

  it('can display loading plans', function () {
    mockEntity.set('options', null);
    var apiUrl = 'http://example.com';
    var renderMarkdown = sinon.spy();
    var getFile = sinon.spy();
    var pluralize = sinon.spy();
    var changeState = sinon.spy();
    var renderer = jsTestUtils.shallowRender(React.createElement(juju.components.EntityContent, {
      apiUrl: apiUrl,
      changeState: changeState,
      entityModel: mockEntity,
      getFile: getFile,
      hasPlans: true,
      plans: null,
      pluralize: pluralize,
      renderMarkdown: renderMarkdown }), true);
    var output = renderer.getRenderOutput();
    var expected = React.createElement(juju.components.Spinner, null);
    assert.deepEqual(output.props.children[1], expected);
  });

  it('can remove plans when none exist', function () {
    mockEntity.set('options', null);
    var apiUrl = 'http://example.com';
    var renderMarkdown = sinon.spy();
    var getFile = sinon.spy();
    var pluralize = sinon.spy();
    var changeState = sinon.spy();
    var renderer = jsTestUtils.shallowRender(React.createElement(juju.components.EntityContent, {
      apiUrl: apiUrl,
      changeState: changeState,
      entityModel: mockEntity,
      getFile: getFile,
      hasPlans: true,
      plans: [],
      pluralize: pluralize,
      renderMarkdown: renderMarkdown }), true);
    var output = renderer.getRenderOutput();
    var expected = undefined;
    assert.deepEqual(output.props.children[1], expected);
  });
});