/*
This file is part of the Juju GUI, which lets users view and manage Juju
environments within a graphical interface (https://launchpad.net/juju-gui).
Copyright (C) 2015 Canonical Ltd.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License version 3, as published by
the Free Software Foundation.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
General Public License for more details.

You should have received a copy of the GNU Affero General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

'use strict';

var _this = this;

var juju = { components: {} }; // eslint-disable-line no-unused-vars

chai.config.includeStack = true;
chai.config.truncateThreshold = 0;

describe('SectionLoadWatcher', () => {
  let SectionLoadWatcher; // eslint-disable-line no-unused-vars
  let EmptyComponent, ErrorComponent, TestComponent;

  beforeAll(done => YUI().use('section-load-watcher', () => {
    done();
  }));

  beforeEach(() => {
    SectionLoadWatcher = juju.components.SectionLoadWatcher;
    TestComponent = React.createClass({
      render: function () {
        return React.createElement('div', { className: 'test' });
      }
    });
    EmptyComponent = React.createClass({
      render: function () {
        return React.createElement('div', { className: 'empty' });
      }
    });
    ErrorComponent = React.createClass({
      render: function () {
        return React.createElement('div', { className: 'error' });
      }
    });
  });

  it('augments a child component', () => {
    const renderer = jsTestUtils.shallowRender(React.createElement(
      SectionLoadWatcher,
      null,
      React.createElement(TestComponent, { ref: 'test1' })
    ), true);
    const instance = renderer.getMountedInstance();
    const output = renderer.getRenderOutput();
    output.props.children[0].props.broadcastStatus('starting');
    assert.equal(instance._childrenStatuses.get('test1'), 'starting');
  });

  it('renders all child components', () => {
    const output = jsTestUtils.shallowRender(React.createElement(
      SectionLoadWatcher,
      null,
      React.createElement(TestComponent, { ref: 'test1' }),
      React.createElement(TestComponent, { ref: 'test2' })
    ));
    assert.equal(output.props.children.length, 2);
    assert.equal(output.props.children[0].type, TestComponent);
    assert.equal(output.props.children[0].ref, 'test1');
    assert.equal(output.props.children[1].type, TestComponent);
    assert.equal(output.props.children[1].ref, 'test2');
  });

  it('renders the empty component when all children report empty', () => {
    const renderer = jsTestUtils.shallowRender(React.createElement(
      SectionLoadWatcher,
      {
        EmptyComponent: React.createElement(EmptyComponent, null) },
      React.createElement(TestComponent, { ref: 'test1' }),
      React.createElement(TestComponent, { ref: 'test2' })
    ), true);
    let output = renderer.getRenderOutput();
    output.props.children[0].props.broadcastStatus('empty');
    output.props.children[1].props.broadcastStatus('empty');
    // Trigger the render again
    output = renderer.getRenderOutput();
    assert.equal(output.props.children.type, EmptyComponent);
  });

  it('renders the error component when all children report an error', () => {
    const renderer = jsTestUtils.shallowRender(React.createElement(
      SectionLoadWatcher,
      {
        ErrorComponent: React.createElement(ErrorComponent, null) },
      React.createElement(TestComponent, { ref: 'test1' }),
      React.createElement(TestComponent, { ref: 'test2' })
    ), true);
    let output = renderer.getRenderOutput();
    output.props.children[0].props.broadcastStatus('error');
    output.props.children[1].props.broadcastStatus('error');
    // Trigger the render again
    output = renderer.getRenderOutput();
    assert.equal(output.props.children.type, ErrorComponent);
  });

  it('renders the children if not all report empty', () => {
    const renderer = jsTestUtils.shallowRender(React.createElement(
      SectionLoadWatcher,
      null,
      React.createElement(TestComponent, { ref: 'test1' }),
      React.createElement(TestComponent, { ref: 'test2' })
    ), true);
    let output = renderer.getRenderOutput();
    output.props.children[0].props.broadcastStatus('empty');
    // Trigger the render again
    output = renderer.getRenderOutput();
    assert.equal(output.props.children.length, 2);
    assert.equal(output.props.children[0].type, TestComponent);
    assert.equal(output.props.children[0].ref, 'test1');
    assert.equal(output.props.children[1].type, TestComponent);
    assert.equal(output.props.children[1].ref, 'test2');
  });

  it('renders the children if not all report error', () => {
    const renderer = jsTestUtils.shallowRender(React.createElement(
      SectionLoadWatcher,
      null,
      React.createElement(TestComponent, { ref: 'test1' }),
      React.createElement(TestComponent, { ref: 'test2' })
    ), true);
    let output = renderer.getRenderOutput();
    output.props.children[0].props.broadcastStatus('error');
    // Trigger the render again
    output = renderer.getRenderOutput();
    assert.equal(output.props.children.length, 2);
    assert.equal(output.props.children[0].type, TestComponent);
    assert.equal(output.props.children[0].ref, 'test1');
    assert.equal(output.props.children[1].type, TestComponent);
    assert.equal(output.props.children[1].ref, 'test2');
  });

  it('throws if a child tries to set an invalid status', () => {
    const renderer = jsTestUtils.shallowRender(React.createElement(
      SectionLoadWatcher,
      null,
      React.createElement(TestComponent, { ref: 'test1' })
    ), true);
    const instance = renderer.getMountedInstance();
    const output = renderer.getRenderOutput();
    assert.throws(output.props.children[0].props.broadcastStatus.bind(instance, 'invalid'), 'Invalid status: "invalid" from ref: test1');
  });

  it('throws if a child is missing a ref', () => {
    assert.throws(jsTestUtils.shallowRender.bind(_this, React.createElement(
      SectionLoadWatcher,
      null,
      React.createElement(TestComponent, null)
    )), 'ref required but not supplied for component');
  });
});