/*
This file is part of the Juju GUI, which lets users view and manage Juju
environments within a graphical interface (https://launchpad.net/juju-gui).
Copyright (C) 2015 Canonical Ltd.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License version 3, as published by
the Free Software Foundation.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
General Public License for more details.

You should have received a copy of the GNU Affero General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

'use strict';

YUI.add('store', function () {

    juju.components.Store = React.createClass({
        propTypes: {
            apiVersion: React.PropTypes.string.isRequired,
            changeState: React.PropTypes.func.isRequired,
            charmstoreURL: React.PropTypes.string.isRequired,
            staticURL: React.PropTypes.string
        },

        /**
          Generate the path for local images on the store page
           @method _generateLocalImagePath
          @param {String} filename The name of the file
          @return {String} A path to the local asset
        */
        _generateLocalImagePath: function (filename) {
            var staticURL = this.props.staticURL || '';
            var basePath = `${ staticURL }/static/gui/build/app/assets/images/store/`;
            return basePath + filename;
        },

        /**
          Generate the URL to a icon for a charm or bundle
           @method _generateIconPath
          @param {String} entityName The name of the charm or bundle
          @return {String} A URL to the entity icon
        */
        _generateIconPath: function (entityName) {
            var charmstoreURL = this.props.charmstoreURL;
            var apiVersion = this.props.apiVersion;
            return `${ charmstoreURL }${ apiVersion }/${ entityName }/icon.svg`;
        },

        /**
          Generate the URL to a diagram for a bundle
           @method _generateDiagramPath
          @param {String} entityName The name of the bundle
          @return {String} A URL to the bundles diagram
        */
        _generateDiagramPath: function (entityName) {
            var charmstoreURL = this.props.charmstoreURL;
            var apiVersion = this.props.apiVersion;
            return `${ charmstoreURL }${ apiVersion }/bundle/${ entityName }/diagram.svg`;
        },

        /**
          Generate the URL to a users page on the storefront
           @method _generateUserPath
          @param {String} username The name of the user
          @return {String} A URL to the users page on storefront
        */
        _generateUserPath: function (username) {
            return `https://jujucharms.com/u/${ username }`;
        },

        /**
          Show the entity details when clicked.
           @method _handleEntityClick
          @param {String} id The entity id.
          @param {Object} e The click event.
        */
        _handleEntityClick: function (e) {
            e.stopPropagation();
            var id = e.target.dataset.entity || e.target.closest('[data-entity]').dataset.entity;
            this.props.changeState({
                sectionC: {
                    component: 'charmbrowser',
                    metadata: {
                        activeComponent: 'entity-details',
                        id: id
                    }
                }
            });
        },

        /**
          Show the search results when clicked.
           @method _handleSearchClick
          @param {String} query The search string.
          @param {Object} e The click event.
        */
        _handleSearchClick: function (e) {
            e.stopPropagation();
            var query = e.target.dataset.query || e.target.closest('[data-query]').dataset.query;
            this.props.changeState({
                sectionC: {
                    component: 'charmbrowser',
                    metadata: {
                        activeComponent: 'search-results',
                        search: query
                    }
                }
            });
        },

        /**
          The content for the write you own section of the page
           @method _writeOurOwnSection
          @return {Object} The contents of the section
        */
        _writeOurOwnSection: function () {
            var doc = 'https://www.jujucharms.com/docs/stable/authors-charm-writing';
            return React.createElement(
                'div',
                { className: 'row row--write-your-own' },
                React.createElement(
                    'div',
                    { className: 'inner-wrapper clearfix' },
                    React.createElement(
                        'div',
                        { className: 'six-col' },
                        React.createElement(
                            'h2',
                            null,
                            'Write a charm and join the ecosystem'
                        ),
                        React.createElement(
                            'p',
                            null,
                            'Creating new charms it easy. Charms can be written in your choice of language and adapting existing scripts is straightforward. You can keep new charms private, or share them back with the community.'
                        ),
                        React.createElement(
                            'p',
                            null,
                            React.createElement(
                                'a',
                                { target: '_blank',
                                    href: doc },
                                'Learn more about writing charms\xA0\u203A'
                            )
                        )
                    )
                )
            );
        },

        /**
          The content for the development tools section of the page
           @method _developmentToolsSection
          @return {Object} The contents of the section
        */
        _developmentToolsSection: function () {
            return React.createElement(
                'div',
                { className: 'row row--border-bottom' },
                React.createElement(
                    'div',
                    { className: 'inner-wrapper' },
                    React.createElement(
                        'h2',
                        null,
                        'Development tools'
                    ),
                    React.createElement(
                        'ul',
                        { className: 'twelve-col no-bullets equal-height' },
                        React.createElement(
                            'li',
                            { className: 'three-col featured-entity' },
                            React.createElement(
                                'span',
                                { className: 'featured-entity__link link',
                                    'data-entity': 'gitlab',
                                    onClick: this._handleEntityClick },
                                React.createElement(
                                    'div',
                                    { className: 'one-col no-margin-bottom' },
                                    React.createElement('img', { src: this._generateIconPath('gitlab'),
                                        alt: '',
                                        className: 'featured-entity__image' })
                                ),
                                React.createElement(
                                    'div',
                                    {
                                        className: 'two-col last-col no-margin-bottom' },
                                    React.createElement(
                                        'h3',
                                        { className: 'featured-entity__title' },
                                        'Gitlab'
                                    ),
                                    React.createElement(
                                        'p',
                                        null,
                                        'by charmers'
                                    )
                                )
                            )
                        ),
                        React.createElement(
                            'li',
                            { className: 'three-col featured-entity' },
                            React.createElement(
                                'span',
                                { className: 'featured-entity__link link',
                                    'data-entity': 'jenkins',
                                    onClick: this._handleEntityClick },
                                React.createElement(
                                    'div',
                                    { className: 'one-col no-margin-bottom' },
                                    React.createElement('img', { src: this._generateIconPath('jenkins'),
                                        alt: '',
                                        className: 'featured-entity__image' })
                                ),
                                React.createElement(
                                    'div',
                                    { className: 'two-col last-col no-margin-bottom' },
                                    React.createElement(
                                        'h3',
                                        { className: 'featured-entity__title' },
                                        'Jenkins'
                                    ),
                                    React.createElement(
                                        'p',
                                        null,
                                        'by Creative'
                                    )
                                )
                            )
                        )
                    )
                )
            );
        },

        /**
          The content for the featured charms and bundles section of the page
           @method _featuredSection
          @return {Object} The contents of the section
        */
        _featuredSection: function () {
            return React.createElement(
                'div',
                { className: 'row equal-height' },
                React.createElement(
                    'div',
                    { className: 'box box--kibana align-center four-col' },
                    React.createElement('img', { src: this._generateLocalImagePath('kabana-promo.png'),
                        alt: '', className: 'box__image' }),
                    React.createElement(
                        'div',
                        { className: 'align-bottom' },
                        React.createElement(
                            'h2',
                            null,
                            'Kibana'
                        ),
                        React.createElement(
                            'span',
                            {
                                'data-entity': 'kibana',
                                onClick: this._handleEntityClick,
                                className: 'button--inline-neutral' },
                            'View the charm'
                        )
                    )
                ),
                React.createElement(
                    'div',
                    { className: 'box box--kubernetes align-center four-col' },
                    React.createElement('img', { src: this._generateLocalImagePath('kubernetes-promo.png'),
                        alt: '', className: 'box__image' }),
                    React.createElement(
                        'div',
                        { className: 'align-bottom' },
                        React.createElement(
                            'h2',
                            null,
                            'Containers'
                        ),
                        React.createElement(
                            'span',
                            {
                                'data-query': 'containers&type=bundle',
                                onClick: this._handleSearchClick,
                                className: 'button--inline-neutral' },
                            'View bundles'
                        )
                    )
                ),
                React.createElement(
                    'div',
                    {
                        className: 'box box--openstack align-center four-col last-col' },
                    React.createElement('img', { src: this._generateLocalImagePath('openstack-promo.png'),
                        alt: '', className: 'box__image' }),
                    React.createElement(
                        'div',
                        { className: 'align-bottom' },
                        React.createElement(
                            'h2',
                            null,
                            'OpenStack'
                        ),
                        React.createElement(
                            'span',
                            {
                                'data-query': 'openstack&type=bundle',
                                onClick: this._handleSearchClick,
                                className: 'button--inline-neutral' },
                            'View bundles'
                        )
                    )
                )
            );
        },

        /**
          Generate contents of the tagged list
           @method _tagsSection
          @return {Object} A generated section listing topics with counts
        */
        _tagsSection: function () {
            var topics = [{ name: 'databases', count: 19 }, { name: 'app-servers', count: 19 }, { name: 'file-servers', count: 16 }, { name: 'monitoring', count: 14 }, { name: 'ops', count: 9 }, { name: 'openstack', count: 51 }, { name: 'applications', count: 75 }, { name: 'misc', count: 63 }, { name: 'network', count: 11 }, { name: 'analytics', count: 7 }, { name: 'apache', count: 38 }, { name: 'security', count: 4 }, { name: 'storage', count: 17 }];
            var list = [];
            topics.forEach(function (topic, index) {
                var url = `&tags=${ topic.name }`;
                var key = `tagItem-${ index }`;
                list.push(React.createElement(
                    'li',
                    { className: 'inline-list__item', key: key },
                    React.createElement(
                        'span',
                        { onClick: this._handleSearchClick,
                            'data-query': url,
                            className: 'link' },
                        topic.name
                    ),
                    React.createElement(
                        'span',
                        { className: 'note' },
                        '(',
                        topic.count,
                        ')'
                    ),
                    ','
                ));
            }, this);
            return React.createElement(
                'div',
                { className: 'eight-col prepend-two align-center' },
                React.createElement(
                    'ul',
                    { className: 'no-bullets inline-list' },
                    list
                )
            );
        },

        /**
          The content for the description of what a charm and bundle are section
          of the page
           @method _charmAndBundleSection
          @return {Object} The contents of the section
        */
        _charmAndBundleSection: function () {
            return React.createElement(
                'div',
                { className: 'row row--charm-and-bundle row--border-bottom' },
                React.createElement(
                    'div',
                    { className: 'inner-wrapper equal-height' },
                    React.createElement(
                        'div',
                        { className: 'six-col box' },
                        React.createElement(
                            'div',
                            { className: 'one-col no-margin-bottom align-center' },
                            React.createElement('img', {
                                src: this._generateLocalImagePath('charm-icon.png'),
                                alt: '' })
                        ),
                        React.createElement(
                            'div',
                            { className: 'five-col no-margin-bottom last-col' },
                            React.createElement(
                                'p',
                                null,
                                'Charms are sets of scripts that simplify the deployment and management tasks of a service. They are regularly reviewed and updated.'
                            ),
                            React.createElement(
                                'span',
                                { onClick: this._handleSearchClick,
                                    'data-query': '&type=charm',
                                    className: 'button--inline-neutral' },
                                'View all the charms'
                            )
                        )
                    ),
                    React.createElement(
                        'div',
                        { className: 'six-col last-col box' },
                        React.createElement(
                            'div',
                            { className: 'one-col no-margin-bottom align-center' },
                            React.createElement('img', {
                                src: this._generateLocalImagePath('bundle-icon.png'),
                                alt: '' })
                        ),
                        React.createElement(
                            'div',
                            { className: 'five-col no-margin-bottom last-col' },
                            React.createElement(
                                'p',
                                null,
                                'Bundles are collections of charms that link applications together, so you can deploy whole chunks of infrastructure in one go.'
                            ),
                            React.createElement(
                                'span',
                                { onClick: this._handleSearchClick,
                                    'data-query': '&type=bundle',
                                    className: 'button--inline-neutral' },
                                'View all the bundles'
                            )
                        )
                    ),
                    this._tagsSection()
                )
            );
        },

        /**
          The content for the operations section of the page
           @method _operationsSection
          @return {Object} The contents of the section
        */
        _operationsSection: function () {
            return React.createElement(
                'div',
                { className: 'row' },
                React.createElement(
                    'div',
                    { className: 'inner-wrapper' },
                    React.createElement(
                        'h2',
                        null,
                        'Operations'
                    ),
                    React.createElement(
                        'div',
                        { className: 'box box--nagios clearfix' },
                        React.createElement(
                            'div',
                            { className: 'six-col no-margin-bottom align-center' },
                            React.createElement('img', {
                                src: this._generateLocalImagePath('nagios-promo.png'),
                                alt: '' })
                        ),
                        React.createElement(
                            'div',
                            { className: 'six-col no-margin-bottom last-col' },
                            React.createElement(
                                'h3',
                                null,
                                'Nagios'
                            ),
                            React.createElement(
                                'p',
                                null,
                                'By ',
                                React.createElement(
                                    'a',
                                    { href: this._generateUserPath('charmers'),
                                        target: '_blank' },
                                    'charmers'
                                )
                            ),
                            React.createElement(
                                'p',
                                null,
                                'Nagios offers complete monitoring, management and alerting of any service from the charm store that is related to it.'
                            ),
                            React.createElement(
                                'span',
                                {
                                    'data-entity': 'nagios',
                                    onClick: this._handleEntityClick,
                                    className: 'button--inline-neutral' },
                                'View the charm'
                            )
                        )
                    ),
                    React.createElement(
                        'ul',
                        { className: 'twelve-col no-bullets equal-height' },
                        React.createElement(
                            'li',
                            { className: 'three-col featured-entity' },
                            React.createElement(
                                'span',
                                { className: 'featured-entity__link link',
                                    'data-entity': 'kibana',
                                    onClick: this._handleEntityClick },
                                React.createElement(
                                    'div',
                                    { className: 'one-col no-margin-bottom' },
                                    React.createElement('img', { src: this._generateIconPath('kibana'),
                                        alt: '',
                                        className: 'featured-entity__image' })
                                ),
                                React.createElement(
                                    'div',
                                    {
                                        className: 'two-col last-col no-margin-bottom' },
                                    React.createElement(
                                        'h3',
                                        { className: 'featured-entity__title' },
                                        'kibana'
                                    ),
                                    React.createElement(
                                        'p',
                                        null,
                                        'by containers'
                                    )
                                )
                            )
                        ),
                        React.createElement(
                            'li',
                            { className: 'three-col featured-entity' },
                            React.createElement(
                                'span',
                                { className: 'featured-entity__link link',
                                    'data-entity': 'logstash',
                                    onClick: this._handleEntityClick },
                                React.createElement(
                                    'div',
                                    { className: 'one-col no-margin-bottom' },
                                    React.createElement('img', {
                                        src: this._generateIconPath('logstash'),
                                        alt: '',
                                        className: 'featured-entity__image' })
                                ),
                                React.createElement(
                                    'div',
                                    {
                                        className: 'two-col last-col no-margin-bottom' },
                                    React.createElement(
                                        'h3',
                                        { className: 'featured-entity__title' },
                                        'logstash'
                                    ),
                                    React.createElement(
                                        'p',
                                        null,
                                        'by containers'
                                    )
                                )
                            )
                        ),
                        React.createElement(
                            'li',
                            { className: 'three-col featured-entity' },
                            React.createElement(
                                'span',
                                { className: 'featured-entity__link link',
                                    'data-entity': 'elasticsearch',
                                    onClick: this._handleEntityClick },
                                React.createElement(
                                    'div',
                                    { className: 'one-col no-margin-bottom' },
                                    React.createElement('img', {
                                        src: this._generateIconPath('elasticsearch'),
                                        alt: '',
                                        className: 'featured-entity__image' })
                                ),
                                React.createElement(
                                    'div',
                                    {
                                        className: 'two-col last-col no-margin-bottom' },
                                    React.createElement(
                                        'h3',
                                        { className: 'featured-entity__title' },
                                        'Elasticsearch'
                                    ),
                                    React.createElement(
                                        'p',
                                        null,
                                        'by onlineservices-charmers'
                                    )
                                )
                            )
                        ),
                        React.createElement(
                            'li',
                            { className: 'three-col last-col featured-entity' },
                            React.createElement(
                                'span',
                                { className: 'featured-entity__link link',
                                    'data-entity': '~canonical-bootstack/prometheus',
                                    onClick: this._handleEntityClick },
                                React.createElement(
                                    'div',
                                    { className: 'one-col no-margin-bottom' },
                                    React.createElement('img', {
                                        src: this._generateIconPath('~canonical-bootstack/prometheus'),
                                        alt: '',
                                        className: 'featured-entity__image' })
                                ),
                                React.createElement(
                                    'div',
                                    {
                                        className: 'two-col last-col no-margin-bottom' },
                                    React.createElement(
                                        'h3',
                                        { className: 'featured-entity__title' },
                                        'Prometheus'
                                    ),
                                    React.createElement(
                                        'p',
                                        null,
                                        'by canonical-bootstack'
                                    )
                                )
                            )
                        ),
                        React.createElement(
                            'li',
                            { className: 'three-col featured-entity' },
                            React.createElement(
                                'a',
                                { className: 'featured-entity__link link',
                                    'data-entity': 'munin',
                                    onClick: this._handleEntityClick },
                                React.createElement(
                                    'div',
                                    {
                                        className: 'one-col no-margin-bottom' },
                                    React.createElement('img', { src: this._generateIconPath('munin'),
                                        alt: '',
                                        className: 'featured-entity__image' })
                                ),
                                React.createElement(
                                    'div',
                                    {
                                        className: 'two-col last-col no-margin-bottom' },
                                    React.createElement(
                                        'h3',
                                        { className: 'featured-entity__title' },
                                        'Munin'
                                    ),
                                    React.createElement(
                                        'p',
                                        null,
                                        'by charmers'
                                    )
                                )
                            )
                        ),
                        React.createElement(
                            'li',
                            { className: 'three-col featured-entity' },
                            React.createElement(
                                'span',
                                { className: 'featured-entity__link link',
                                    'data-entity': 'rsyslog',
                                    onClick: this._handleEntityClick },
                                React.createElement(
                                    'div',
                                    {
                                        className: 'one-col no-margin-bottom' },
                                    React.createElement('img', { src: this._generateIconPath('rsyslog'),
                                        alt: '',
                                        className: 'featured-entity__image' })
                                ),
                                React.createElement(
                                    'div',
                                    { className: 'two-col last-col no-margin-bottom' },
                                    React.createElement(
                                        'h3',
                                        { className: 'featured-entity__title' },
                                        'Rsyslog'
                                    ),
                                    React.createElement(
                                        'p',
                                        null,
                                        'by charmers'
                                    )
                                )
                            )
                        ),
                        React.createElement(
                            'li',
                            { className: 'three-col featured-entity' },
                            React.createElement(
                                'span',
                                { className: 'featured-entity__link link',
                                    'data-entity': 'zabbix-server',
                                    onClick: this._handleEntityClick },
                                React.createElement(
                                    'div',
                                    { className: 'one-col no-margin-bottom' },
                                    React.createElement('img', {
                                        src: this._generateIconPath('zabbix-server'),
                                        alt: '',
                                        className: 'featured-entity__image control-size'
                                    })
                                ),
                                React.createElement(
                                    'div',
                                    { className: 'two-col last-col no-margin-bottom' },
                                    React.createElement(
                                        'h3',
                                        { className: 'featured-entity__title' },
                                        'Zabbix Server'
                                    ),
                                    React.createElement(
                                        'p',
                                        null,
                                        'by charmers'
                                    )
                                )
                            )
                        ),
                        React.createElement(
                            'li',
                            { className: 'three-col last-col featured-entity' },
                            React.createElement(
                                'span',
                                { className: 'featured-entity__link link',
                                    'data-entity': '~ricardokirkner/sentry',
                                    onClick: this._handleEntityClick },
                                React.createElement(
                                    'div',
                                    { className: 'one-col no-margin-bottom' },
                                    React.createElement('img', { src: this._generateIconPath('~ricardokirkner/sentry'),
                                        alt: '',
                                        className: 'featured-entity__image control-size'
                                    })
                                ),
                                React.createElement(
                                    'div',
                                    {
                                        className: 'two-col last-col no-margin-bottom' },
                                    React.createElement(
                                        'h3',
                                        { className: 'featured-entity__title' },
                                        'Sentry'
                                    ),
                                    React.createElement(
                                        'p',
                                        null,
                                        'by ricardokirkner'
                                    )
                                )
                            )
                        )
                    ),
                    React.createElement(
                        'p',
                        { className: 'intro' },
                        React.createElement(
                            'span',
                            { onClick: this._handleSearchClick,
                                'data-query': '&tags=ops',
                                className: 'link' },
                            'View all operations\xA0\u203A'
                        )
                    )
                )
            );
        },

        /**
          The content for the bigdata section of the page
           @method _bigDataSection
          @return {Object} The contents of the section
        */
        _bigDataSection: function () {
            return React.createElement(
                'div',
                { className: 'row row--bigdata' },
                React.createElement(
                    'div',
                    { className: 'inner-wrapper clearfix' },
                    React.createElement(
                        'div',
                        { className: 'six-col' },
                        React.createElement(
                            'h2',
                            null,
                            'Big data charms and bundles'
                        ),
                        React.createElement(
                            'p',
                            null,
                            'Spend time testing, evaluating, and using your Big Data solutions to benefit your business by using one of Juju\u2019s pre-configured bundles or expertly created charms.'
                        ),
                        React.createElement(
                            'p',
                            null,
                            React.createElement(
                                'span',
                                { onClick: this._handleSearchClick,
                                    'data-query': '&type=bigdata',
                                    className: 'link' },
                                'Big data charms and bundles\xA0\u203A'
                            )
                        )
                    ),
                    React.createElement(
                        'div',
                        { className: 'four-col last-col prepend-one append-one' },
                        React.createElement('img', {
                            src: this._generateLocalImagePath('hadoop-elephant.png'),
                            className: 'for-small',
                            alt: 'Hadoop logo' })
                    )
                )
            );
        },

        /**
          The content for the analytics section of the page
           @method _analyticsSection
          @return {Object} The contents of the section
        */
        _analyticsSection: function () {
            return React.createElement(
                'div',
                { className: 'row row--border-bottom' },
                React.createElement(
                    'div',
                    { className: 'inner-wrapper' },
                    React.createElement(
                        'h2',
                        null,
                        'Analytics'
                    ),
                    React.createElement(
                        'div',
                        { className: 'box box--realtime-syslog-analytics clearfix' },
                        React.createElement(
                            'div',
                            { className: 'six-col no-margin-bottom align-center' },
                            React.createElement('object', {
                                wmode: 'transparent',
                                width: '100%',
                                type: 'image/svg+xml',
                                data: this._generateDiagramPath('realtime-syslog-analytics') })
                        ),
                        React.createElement(
                            'div',
                            { className: 'six-col no-margin-bottom last-col' },
                            React.createElement(
                                'h3',
                                null,
                                'Realtime Syslog Analytics'
                            ),
                            React.createElement(
                                'p',
                                null,
                                'By ',
                                React.createElement(
                                    'a',
                                    { href: this._generateUserPath('bigdata-charmers'),
                                        target: '_blank' },
                                    'bigdata-charmers'
                                )
                            ),
                            React.createElement(
                                'p',
                                null,
                                'This bundle provides a big data environment for analysing syslog events. Built around Apache Hadoop components, it offers a repeatable and reliable way to setup complex software across multiple substrates.'
                            ),
                            React.createElement(
                                'span',
                                { onClick: this._handleEntityClick,
                                    'data-entity': 'realtime-syslog-analytics',
                                    className: 'button--inline-neutral' },
                                'View the bundle'
                            )
                        )
                    ),
                    React.createElement(
                        'ul',
                        { className: 'twelve-col no-bullets equal-height' },
                        React.createElement(
                            'li',
                            { className: 'three-col featured-entity' },
                            React.createElement(
                                'span',
                                { className: 'featured-entity__link link',
                                    'data-entity': 'apache-analytics-sql',
                                    onClick: this._handleEntityClick },
                                React.createElement(
                                    'ul',
                                    { className: 'featured-entity__image-list one-col' },
                                    React.createElement(
                                        'li',
                                        { className: 'featured-entity__image-list-item' },
                                        React.createElement('img', { src: this._generateIconPath('apache-hadoop-slave'),
                                            alt: '',
                                            className: 'featured-entity__image' })
                                    ),
                                    React.createElement(
                                        'li',
                                        { className: 'featured-entity__image-list-item' },
                                        React.createElement('img', { src: this._generateIconPath('apache-hive'),
                                            alt: '',
                                            className: 'featured-entity__image' })
                                    ),
                                    React.createElement(
                                        'li',
                                        { className: 'featured-entity__image-list-item' },
                                        React.createElement('img', { src: this._generateIconPath('mariadb'),
                                            alt: '',
                                            className: 'featured-entity__image' })
                                    ),
                                    React.createElement(
                                        'li',
                                        { className: 'featured-entity__image-list-item' },
                                        '+3'
                                    )
                                ),
                                React.createElement(
                                    'div',
                                    { className: 'two-col last-col no-margin-bottom' },
                                    React.createElement(
                                        'h3',
                                        { className: 'featured-entity__title' },
                                        'Apache analytics SQL'
                                    ),
                                    React.createElement(
                                        'p',
                                        null,
                                        'by bigdata-charmers'
                                    )
                                )
                            )
                        ),
                        React.createElement(
                            'li',
                            { className: 'three-col featured-entity' },
                            React.createElement(
                                'span',
                                { className: 'featured-entity__link link',
                                    'data-entity': 'apache-spark',
                                    onClick: this._handleEntityClick },
                                React.createElement(
                                    'div',
                                    { className: 'one-col no-margin-bottom' },
                                    React.createElement('img', {
                                        src: this._generateIconPath('apache-spark'),
                                        alt: '',
                                        className: 'featured-entity__image' })
                                ),
                                React.createElement(
                                    'div',
                                    {
                                        className: 'two-col last-col no-margin-bottom' },
                                    React.createElement(
                                        'h3',
                                        { className: 'featured-entity__title' },
                                        'Apache Spark'
                                    ),
                                    React.createElement(
                                        'p',
                                        null,
                                        'by bigdata-charmers'
                                    )
                                )
                            )
                        ),
                        React.createElement(
                            'li',
                            { className: 'three-col featured-entity' },
                            React.createElement(
                                'span',
                                { className: 'featured-entity__link link',
                                    'data-entity': 'apache-zeppelin',
                                    onClick: this._handleEntityClick },
                                React.createElement(
                                    'div',
                                    { className: 'one-col no-margin-bottom' },
                                    React.createElement('img', {
                                        src: this._generateIconPath('apache-zeppelin'),
                                        alt: '',
                                        className: 'featured-entity__image' })
                                ),
                                React.createElement(
                                    'div',
                                    { className: 'two-col last-col no-margin-bottom' },
                                    React.createElement(
                                        'h3',
                                        { className: 'featured-entity__title' },
                                        'Apache Zeppelin'
                                    ),
                                    React.createElement(
                                        'p',
                                        null,
                                        'by bigdata-charmers'
                                    )
                                )
                            )
                        ),
                        React.createElement(
                            'li',
                            { className: 'three-col last-col featured-entity' },
                            React.createElement(
                                'span',
                                { className: 'featured-entity__link link',
                                    'data-entity': 'elasticsearch-cluster',
                                    onClick: this._handleEntityClick },
                                React.createElement(
                                    'ul',
                                    { className: 'featured-entity__image-list one-col' },
                                    React.createElement(
                                        'li',
                                        { className: 'featured-entity__image-list-item' },
                                        React.createElement('img', {
                                            src: this._generateIconPath('elasticsearch'),
                                            alt: '',
                                            className: 'featured-entity__image' })
                                    ),
                                    React.createElement(
                                        'li',
                                        { className: 'featured-entity__image-list-item' },
                                        React.createElement('img', { src: this._generateIconPath('kibana'),
                                            alt: '',
                                            className: 'featured-entity__image' })
                                    )
                                ),
                                React.createElement(
                                    'div',
                                    { className: 'two-col last-col no-margin-bottom' },
                                    React.createElement(
                                        'h3',
                                        { className: 'featured-entity__title' },
                                        'Elasticsearch cluster'
                                    ),
                                    React.createElement(
                                        'p',
                                        null,
                                        'by charmers'
                                    )
                                )
                            )
                        ),
                        React.createElement(
                            'li',
                            { className: 'three-col featured-entity' },
                            React.createElement(
                                'span',
                                { className: 'featured-entity__link link',
                                    'data-entity': '~containers/elk-stack',
                                    onClick: this._handleEntityClick },
                                React.createElement(
                                    'ul',
                                    { className: 'featured-entity__image-list one-col' },
                                    React.createElement(
                                        'li',
                                        { className: 'featured-entity__image-list-item' },
                                        React.createElement('img', { src: this._generateIconPath('zulu8'),
                                            alt: '',
                                            className: 'featured-entity__image' })
                                    ),
                                    React.createElement(
                                        'li',
                                        { className: 'featured-entity__image-list-item' },
                                        React.createElement('img', {
                                            src: this._generateIconPath('elasticsearch'),
                                            alt: '',
                                            className: 'featured-entity__image' })
                                    ),
                                    React.createElement(
                                        'li',
                                        { className: 'featured-entity__image-list-item' },
                                        React.createElement('img', {
                                            src: this._generateIconPath('~containers/logstash'),
                                            alt: '',
                                            className: 'featured-entity__image' })
                                    ),
                                    React.createElement(
                                        'li',
                                        { className: 'featured-entity__image-list-item' },
                                        React.createElement('img', { src: this._generateIconPath('kibana'),
                                            alt: '',
                                            className: 'featured-entity__image' })
                                    )
                                ),
                                React.createElement(
                                    'div',
                                    { className: 'two-col last-col no-margin-bottom' },
                                    React.createElement(
                                        'h3',
                                        { className: 'featured-entity__title' },
                                        'elk stack'
                                    ),
                                    React.createElement(
                                        'p',
                                        null,
                                        'by containers'
                                    )
                                )
                            )
                        )
                    ),
                    React.createElement(
                        'p',
                        { className: 'intro' },
                        React.createElement(
                            'span',
                            { onClick: this._handleSearchClick,
                                'data-query': '&tags=analytics',
                                className: 'link' },
                            'View all analytics\xA0\u203A'
                        )
                    )
                )
            );
        },

        /**
          The content for the databases section of the page
           @method _databasesSection
          @return {Object} The contents of the section
        */
        _databasesSection: function () {
            return React.createElement(
                'div',
                { className: 'row row--border-bottom' },
                React.createElement(
                    'div',
                    { className: 'inner-wrapper' },
                    React.createElement(
                        'h2',
                        null,
                        'Databases'
                    ),
                    React.createElement(
                        'div',
                        { className: 'box box--mysql clearfix' },
                        React.createElement(
                            'div',
                            { className: 'six-col no-margin-bottom align-center' },
                            React.createElement('img', {
                                src: this._generateLocalImagePath('mysql-promo.png'),
                                alt: '' })
                        ),
                        React.createElement(
                            'div',
                            { className: 'six-col no-margin-bottom last-col' },
                            React.createElement(
                                'h3',
                                null,
                                'MySQL'
                            ),
                            React.createElement(
                                'p',
                                null,
                                'By ',
                                React.createElement(
                                    'a',
                                    { href: this._generateUserPath('mysql-charmers'),
                                        target: '_blank' },
                                    'mysql-charmers'
                                )
                            ),
                            React.createElement(
                                'p',
                                null,
                                'MySQL is a fast, stable and true multi-user, multi-threaded SQL database server. Its main goals are speed, robustness and ease of use.'
                            ),
                            React.createElement(
                                'span',
                                { onClick: this._handleEntityClick,
                                    'data-entity': 'mysql',
                                    className: 'button--inline-neutral' },
                                'View the charm'
                            )
                        )
                    ),
                    React.createElement(
                        'ul',
                        { className: 'twelve-col no-bullets equal-height' },
                        React.createElement(
                            'li',
                            { className: 'three-col featured-entity' },
                            React.createElement(
                                'span',
                                { className: 'featured-entity__link link',
                                    'data-entity': 'cassandra',
                                    onClick: this._handleEntityClick },
                                React.createElement(
                                    'div',
                                    { className: 'one-col no-margin-bottom' },
                                    React.createElement('img', { src: this._generateIconPath('cassandra'),
                                        alt: '',
                                        className: 'featured-entity__image' })
                                ),
                                React.createElement(
                                    'div',
                                    { className: 'two-col last-col no-margin-bottom' },
                                    React.createElement(
                                        'h3',
                                        { className: 'featured-entity__title' },
                                        'Cassandra'
                                    ),
                                    React.createElement(
                                        'p',
                                        null,
                                        'by cassandra-charmers'
                                    )
                                )
                            )
                        ),
                        React.createElement(
                            'li',
                            { className: 'three-col featured-entity' },
                            React.createElement(
                                'span',
                                { className: 'featured-entity__link link',
                                    'data-entity': 'mariadb',
                                    onClick: this._handleEntityClick },
                                React.createElement(
                                    'div',
                                    {
                                        className: 'one-col no-margin-bottom' },
                                    React.createElement('img', {
                                        src: this._generateIconPath('mariadb'),
                                        alt: '',
                                        className: 'featured-entity__image' })
                                ),
                                React.createElement(
                                    'div',
                                    {
                                        className: 'two-col last-col no-margin-bottom' },
                                    React.createElement(
                                        'h3',
                                        { className: 'featured-entity__title' },
                                        'Mariadb'
                                    ),
                                    React.createElement(
                                        'p',
                                        null,
                                        'by mariadb-charmers'
                                    )
                                )
                            )
                        ),
                        React.createElement(
                            'li',
                            { className: 'three-col featured-entity' },
                            React.createElement(
                                'span',
                                { className: 'featured-entity__link link',
                                    'data-entity': 'mongodb',
                                    onClick: this._handleEntityClick },
                                React.createElement(
                                    'div',
                                    { className: 'one-col no-margin-bottom' },
                                    React.createElement('img', { src: this._generateIconPath('mongodb'),
                                        alt: '',
                                        className: 'featured-entity__image' })
                                ),
                                React.createElement(
                                    'div',
                                    { className: 'two-col last-col no-margin-bottom' },
                                    React.createElement(
                                        'h3',
                                        { className: 'featured-entity__title' },
                                        'Mongodb'
                                    ),
                                    React.createElement(
                                        'p',
                                        null,
                                        'by charmers'
                                    )
                                )
                            )
                        ),
                        React.createElement(
                            'li',
                            { className: 'three-col last-col featured-entity' },
                            React.createElement(
                                'span',
                                { className: 'featured-entity__link link',
                                    'data-entity': 'redis',
                                    onClick: this._handleEntityClick },
                                React.createElement(
                                    'div',
                                    { className: 'one-col no-margin-bottom' },
                                    React.createElement('img', {
                                        src: this._generateIconPath('redis'),
                                        alt: '',
                                        className: 'featured-entity__image' })
                                ),
                                React.createElement(
                                    'div',
                                    { className: 'two-col last-col no-margin-bottom' },
                                    React.createElement(
                                        'h3',
                                        { className: 'featured-entity__title' },
                                        'Redis'
                                    ),
                                    React.createElement(
                                        'p',
                                        null,
                                        'by charmers'
                                    )
                                )
                            )
                        ),
                        React.createElement(
                            'li',
                            { className: 'three-col last-col featured-entity' },
                            React.createElement(
                                'span',
                                { className: 'featured-entity__link link',
                                    'data-entity': 'postgresql',
                                    onClick: this._handleEntityClick },
                                React.createElement(
                                    'div',
                                    {
                                        className: 'one-col no-margin-bottom' },
                                    React.createElement('img', { src: this._generateIconPath('postgresql'),
                                        alt: '',
                                        className: 'featured-entity__image' })
                                ),
                                React.createElement(
                                    'div',
                                    { className: 'two-col last-col no-margin-bottom' },
                                    React.createElement(
                                        'h3',
                                        { className: 'featured-entity__title' },
                                        'Postgresql'
                                    ),
                                    React.createElement(
                                        'p',
                                        null,
                                        'by postgresql-charmers'
                                    )
                                )
                            )
                        )
                    ),
                    React.createElement(
                        'p',
                        { className: 'intro' },
                        React.createElement(
                            'span',
                            { onClick: this._handleSearchClick,
                                'data-query': '&tags=databases',
                                className: 'link' },
                            'View all databases\xA0\u203A'
                        )
                    )
                )
            );
        },

        render: function () {
            return React.createElement(
                'div',
                { className: 'store' },
                this._featuredSection(),
                this._charmAndBundleSection(),
                this._operationsSection(),
                this._bigDataSection(),
                this._analyticsSection(),
                this._databasesSection(),
                this._developmentToolsSection(),
                this._writeOurOwnSection()
            );
        }
    });
}, '0.1.0', { requires: [] });