/*
This file is part of the Juju GUI, which lets users view and manage Juju
environments within a graphical interface (https://launchpad.net/juju-gui).
Copyright (C) 2016 Canonical Ltd.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License version 3, as published by
the Free Software Foundation.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
General Public License for more details.

You should have received a copy of the GNU Affero General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

'use strict';

var juju = { components: {} }; // eslint-disable-line no-unused-vars

describe('UserProfileEntityList', () => {
  var charmstore, charms, bundles, users;

  beforeAll(done => {
    // By loading this file it adds the component to the juju components.
    YUI().use('user-profile-entity-list', () => {
      done();
    });
  });

  beforeEach(() => {
    var list = sinon.stub();
    var charm = jsTestUtils.makeEntity().toEntity();
    charm.series = [charm.series];
    charms = [charm];
    var bundle = jsTestUtils.makeEntity(true).toEntity();
    bundle.series = [bundle.series];
    bundles = [bundle];
    list.withArgs('test-owner', sinon.match.any, 'charm').callsArgWith(1, null, charms);
    list.withArgs('test-owner', sinon.match.any, 'bundle').callsArgWith(1, null, bundles);
    charmstore = {
      list: list,
      url: 'example.com/9'
    };
    users = { charmstore: {
        user: 'test-owner',
        usernameDisplay: 'test owner'
      } };
  });

  it('renders the empty state', () => {
    var component = jsTestUtils.shallowRender(React.createElement(juju.components.UserProfileEntityList, {
      changeState: sinon.stub(),
      charmstore: {},
      getDiagramURL: sinon.stub(),
      type: 'charm',
      user: users.charmstore,
      users: users }), true);
    var output = component.getRenderOutput();
    assert.equal(output, null);
  });

  it('displays loading spinners for charms and bundles', () => {
    charmstore.list = sinon.stub();
    var type = 'charm';
    var component = jsTestUtils.shallowRender(React.createElement(juju.components.UserProfileEntityList, {
      changeState: sinon.stub(),
      charmstore: charmstore,
      getDiagramURL: sinon.stub(),
      type: type,
      user: users.charmstore,
      users: users }), true);
    var output = component.getRenderOutput();
    assert.deepEqual(output, React.createElement(
      'div',
      { className: 'user-profile__charm-list twelve-col' },
      React.createElement(juju.components.Spinner, null)
    ));
  });

  it('renders a list of charms', () => {
    var changeState = sinon.stub();
    var type = 'charm';
    var component = jsTestUtils.shallowRender(React.createElement(juju.components.UserProfileEntityList, {
      changeState: changeState,
      charmstore: charmstore,
      getDiagramURL: sinon.stub(),
      type: type,
      user: users.charmstore,
      users: users }), true);
    var output = component.getRenderOutput();
    var expected = React.createElement(
      'div',
      { className: 'user-profile__charm-list' },
      React.createElement(
        'div',
        { className: 'user-profile__header twelve-col no-margin-bottom' },
        'Charms',
        React.createElement(
          'span',
          { className: 'user-profile__size' },
          '(',
          1,
          ')'
        )
      ),
      React.createElement(
        'ul',
        { className: 'user-profile__list twelve-col' },
        React.createElement(
          'li',
          { className: 'user-profile__list-header twelve-col' },
          React.createElement(
            'span',
            { className: 'user-profile__list-col three-col' },
            'Name'
          ),
          React.createElement(
            'span',
            { className: 'user-profile__list-col seven-col' },
            'Series'
          ),
          React.createElement(
            'span',
            { className: 'user-profile__list-col two-col last-col' },
            'Owner'
          )
        ),
        [React.createElement(
          juju.components.UserProfileEntity,
          {
            changeState: changeState,
            entity: charms[0],
            key: 'cs:django',
            type: 'charm' },
          React.createElement(
            'span',
            { className: 'user-profile__list-col three-col ' + 'user-profile__list-name' },
            'django',
            React.createElement(
              'ul',
              { className: 'user-profile__list-tags' },
              [React.createElement(
                'li',
                { className: 'user-profile__comma-item',
                  key: 'cs:django-database' },
                'database'
              )]
            )
          ),
          React.createElement(
            'span',
            { className: 'user-profile__list-col four-col' },
            React.createElement(
              'ul',
              { className: 'user-profile__list-series' },
              [React.createElement(
                'li',
                { className: 'user-profile__comma-item',
                  key: 'cs:django-trusty' },
                'trusty'
              )]
            )
          ),
          React.createElement(
            'span',
            { className: 'user-profile__list-col one-col ' + 'user-profile__list-icons' },
            React.createElement('img', { className: 'user-profile__list-icon',
              src: 'example.com/9/django/icon.svg',
              title: 'django' })
          ),
          React.createElement(
            'span',
            { className: 'user-profile__list-col two-col ' + 'prepend-two last-col' },
            'test-owner'
          )
        )]
      )
    );
    assert.deepEqual(output, expected);
  });

  it('renders a list of bundles', () => {
    var changeState = sinon.stub();
    var getDiagramURL = sinon.stub();
    var component = jsTestUtils.shallowRender(React.createElement(juju.components.UserProfileEntityList, {
      changeState: changeState,
      charmstore: charmstore,
      getDiagramURL: getDiagramURL,
      type: 'bundle',
      user: users.charmstore,
      users: users }), true);
    var output = component.getRenderOutput();
    var expected = React.createElement(
      'div',
      { className: 'user-profile__bundle-list' },
      React.createElement(
        'div',
        { className: 'user-profile__header twelve-col no-margin-bottom' },
        'Bundles',
        React.createElement(
          'span',
          { className: 'user-profile__size' },
          '(',
          1,
          ')'
        )
      ),
      React.createElement(
        'ul',
        { className: 'user-profile__list twelve-col' },
        React.createElement(
          'li',
          { className: 'user-profile__list-header twelve-col' },
          React.createElement(
            'span',
            { className: 'user-profile__list-col five-col' },
            'Name'
          ),
          React.createElement(
            'span',
            { className: 'user-profile__list-col three-col user-profile__list-icons' },
            'Charms'
          ),
          React.createElement(
            'span',
            { className: 'user-profile__list-col one-col prepend-one' },
            'Units'
          ),
          React.createElement(
            'span',
            { className: 'user-profile__list-col two-col last-col' },
            'Owner'
          )
        ),
        [React.createElement(
          juju.components.UserProfileEntity,
          {
            changeState: changeState,
            entity: bundles[0],
            getDiagramURL: getDiagramURL,
            key: 'django-cluster',
            type: 'bundle' },
          React.createElement(
            'span',
            { className: 'user-profile__list-col five-col ' + 'user-profile__list-name' },
            'django-cluster',
            React.createElement(
              'ul',
              { className: 'user-profile__list-tags' },
              [React.createElement(
                'li',
                { className: 'user-profile__comma-item',
                  key: 'django-cluster-database' },
                'database'
              )]
            )
          ),
          React.createElement(
            'span',
            { className: 'user-profile__list-col three-col ' + 'user-profile__list-icons' },
            React.createElement('img', { className: 'user-profile__list-icon',
              key: 'icon-0-gunicorn',
              src: 'example.com/9/gunicorn/icon.svg',
              title: 'gunicorn' }),
            React.createElement('img', { className: 'user-profile__list-icon',
              key: 'icon-1-django',
              src: 'example.com/9/django/icon.svg',
              title: 'django' })
          ),
          React.createElement(
            'span',
            { className: 'user-profile__list-col one-col prepend-one' },
            5
          ),
          React.createElement(
            'span',
            { className: 'user-profile__list-col two-col last-col' },
            'test-owner'
          )
        )]
      )
    );
    assert.deepEqual(output, expected);
  });

  it('requests charms and updates state', () => {
    var component = jsTestUtils.shallowRender(React.createElement(juju.components.UserProfileEntityList, {
      changeState: sinon.stub(),
      charmstore: charmstore,
      getDiagramURL: sinon.stub(),
      type: 'charm',
      user: users.charmstore,
      users: users }), true);
    var instance = component.getMountedInstance();
    assert.equal(charmstore.list.callCount, 1, 'charmstore list not called');
    assert.equal(charmstore.list.args[0][0], 'test-owner', 'username not passed to list request');
    assert.deepEqual(instance.state.entityList, charms, 'callback does not properly set entity state');
  });

  it('requests bundles and updates state', () => {
    var component = jsTestUtils.shallowRender(React.createElement(juju.components.UserProfileEntityList, {
      changeState: sinon.stub(),
      charmstore: charmstore,
      getDiagramURL: sinon.stub(),
      type: 'bundle',
      user: users.charmstore,
      users: users }), true);
    var instance = component.getMountedInstance();
    assert.equal(charmstore.list.callCount, 1, 'charmstore list not called');
    assert.equal(charmstore.list.args[0][0], 'test-owner', 'username not passed to list request');
    assert.deepEqual(instance.state.entityList, bundles, 'callback does not properly set entity state');
  });

  it('will abort the requests when unmounting', function () {
    var charmstoreAbort = sinon.stub();
    charmstore.list = sinon.stub().returns({ abort: charmstoreAbort });
    var renderer = jsTestUtils.shallowRender(React.createElement(juju.components.UserProfileEntityList, {
      changeState: sinon.stub(),
      charmstore: charmstore,
      getDiagramURL: sinon.stub(),
      type: 'charm',
      user: users.charmstore,
      users: users }), true);
    renderer.unmount();
    assert.equal(charmstoreAbort.callCount, 1);
  });

  it('gets the entity data when the user authenticates', () => {
    var list = sinon.stub();
    var charmstore = { list: list };
    var component = jsTestUtils.shallowRender(React.createElement(juju.components.UserProfileEntityList, {
      changeState: sinon.stub(),
      charmstore: charmstore,
      getDiagramURL: sinon.stub(),
      type: 'charm',
      users: {},
      user: users.charmstore }), true);
    assert.equal(list.callCount, 0);
    component.render(React.createElement(juju.components.UserProfileEntityList, {
      changeState: sinon.stub(),
      charmstore: charmstore,
      getDiagramURL: sinon.stub(),
      type: 'charm',
      users: users,
      user: users.charmstore }));
    assert.equal(list.callCount, 1);
  });

  it('broadcasts starting status', function () {
    var broadcastStatus = sinon.stub();
    jsTestUtils.shallowRender(React.createElement(juju.components.UserProfileEntityList, {
      broadcastStatus: broadcastStatus,
      changeState: sinon.stub(),
      charmstore: charmstore,
      getDiagramURL: sinon.stub(),
      type: 'charm',
      users: users,
      user: users.charmstore }));
    assert.equal(broadcastStatus.args[0][0], 'starting');
  });

  it('broadcasts ok status', function () {
    var broadcastStatus = sinon.stub();
    jsTestUtils.shallowRender(React.createElement(juju.components.UserProfileEntityList, {
      broadcastStatus: broadcastStatus,
      changeState: sinon.stub(),
      charmstore: charmstore,
      getDiagramURL: sinon.stub(),
      type: 'charm',
      users: users,
      user: users.charmstore }));
    assert.equal(broadcastStatus.args[1][0], 'ok');
  });

  it('broadcasts empty status', function () {
    charmstore.list = sinon.stub().callsArgWith(1, null, []);
    var broadcastStatus = sinon.stub();
    jsTestUtils.shallowRender(React.createElement(juju.components.UserProfileEntityList, {
      broadcastStatus: broadcastStatus,
      changeState: sinon.stub(),
      charmstore: charmstore,
      getDiagramURL: sinon.stub(),
      type: 'charm',
      users: users,
      user: users.charmstore }));
    assert.equal(broadcastStatus.args[1][0], 'empty');
  });

  it('broadcasts error status', function () {
    charmstore.list = sinon.stub().callsArgWith(1, 'error', null);
    var broadcastStatus = sinon.stub();
    jsTestUtils.shallowRender(React.createElement(juju.components.UserProfileEntityList, {
      broadcastStatus: broadcastStatus,
      changeState: sinon.stub(),
      charmstore: charmstore,
      getDiagramURL: sinon.stub(),
      type: 'charm',
      users: users,
      user: users.charmstore }));
    assert.equal(broadcastStatus.args[1][0], 'error');
  });
});